(*
 * Evaluator.
 *
 * ----------------------------------------------------------------
 *
 * @begin[license]
 * Copyright (C) 2000 Jason Hickey, Caltech
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * Author: Jason Hickey
 * @email{jyh@cs.caltech.edu}
 * @end[license]
 *)

open Debug
open Symbol
open Fc_ast_type
open Fc_ast_util

(************************************************************************
 * TYPES AND EXCEPTIONS
 ************************************************************************)

(*
 * Values.
 *)
type pointer =
   PointerNil
 | PointerArray of evalue array * int
 | PointerString of string * int
 | PointerVar of env_var

and evalue =
   Void
 | Bool of bool
 | Char of char
 | Int of int
 | Float of float
 | Function of (pos -> evalue list -> evalue)
 | Struct of fields
 | Pointer of pointer

and fields = (symbol * env_var) list

(*
 * Environment.
 *)
and env_var =
   { mutable var_value : evalue;
     var_type : ty
   }

and env =
   { env_types : ty SymbolTable.t;
     env_vars : env_var SymbolTable.t
   }

(*
 * Exceptions.
 *)
type eval_exn =
   UnboundType of symbol
 | UnboundVar of symbol
 | UnboundField of symbol
 | NilPointer
 | TypeError of string
 | SubscriptOutOfBounds of int
 | DivideByZero
 | NotImplemented of string
 | NotAnInt
 | NotAnLValue
 | NotAScalar
 | NotAPointer
 | ArityMismatch

exception EvalException  of pos * eval_exn
exception Return         of pos * evalue
exception Break          of pos

(************************************************************************
 * UTILITIES
 ************************************************************************)

(*
 * Unary arithmetic.
 *)
let succ_char c =
   Char.chr ((succ (Char.code c)) mod 256)

let pred_char c =
   Char.chr ((pred (Char.code c)) mod 256)

let succ_float x =
   x +. 1.0

let pred_float x =
   x -. 1.0

(************************************************************************
 * PRINTING
 ************************************************************************)

(*
 * A separated list.
 *)
let rec print_sep_array separator printer vals =
   let len = Array.length vals in
      if len = 0 then
         ()
      else
         let len' = pred len in
         let rec loop i =
            if i = len' then
               printer vals.(i)
            else
               begin
                  printer vals.(i);
                  Format.print_string separator;
                  Format.print_space ();
                  loop (succ i)
               end
         in
            loop 0

(*
 * Print a value.
 *)
let rec print_value = function
   Void ->
      Format.print_string "()"
 | Bool b ->
      Format.print_bool b
 | Char c ->
      Format.print_string ("'" ^ Char.escaped c ^ "'")
 | Int i ->
      Format.print_int i
 | Float x ->
      Format.print_float x
 | Function _ ->
      Format.print_string "<fun>"
 | Pointer _ ->
      Format.print_string "<pointer>"
 | Struct fields ->
      Format.open_hvbox 0;
      Format.open_hvbox tabstop;
      Format.print_string "{";
      List.iter (fun (sym, { var_value = v; var_type = ty }) ->
            Format.print_space ();
            Format.open_hvbox tabstop;
            print_type ty;
            Format.print_space ();
            print_symbol sym;
            Format.print_string " =";
            Format.print_space ();
            print_value v;
            Format.print_string ";";
            Format.close_box ()) fields;
      Format.close_box ();
      Format.print_space ();
      Format.print_string "}";
      Format.close_box ()

(************************************************************************
 * ENVIRONMENT
 ************************************************************************)

(*
 * Type environment.
 *)
let lookup_type { env_types = tenv } pos v =
   try SymbolTable.find tenv v with
      Not_found ->
         raise (EvalException (pos, UnboundType v))

let add_type { env_types = tenv; env_vars = venv } v v' =
   { env_types = SymbolTable.add tenv v v'; env_vars = venv }

(*
 * Variable environment.
 *)
let lookup_var { env_vars = venv } pos v =
   try SymbolTable.find venv v with
      Not_found ->
         raise (EvalException (pos, UnboundVar v))

let set_var env pos v v' =
   try
      let v = SymbolTable.find env.env_vars v in
         v.var_value <- v';
         env
   with
      Not_found ->
         raise (EvalException (pos, UnboundVar v))

let add_var { env_types = tenv; env_vars = venv } v v' ty =
   { env_types = tenv;
     env_vars = SymbolTable.add venv v { var_value = v'; var_type = ty }
   }

(*
 * Structure lookup.
 *)
let lookup_struct pos s v =
   try List.assoc v s with
      Not_found ->
         raise (EvalException (pos, UnboundField v))

(************************************************************************
 * COERCIONS
 ************************************************************************)

(*
 * Coerce to a bool type.
 *)
let coerce_bool pos = function
   Bool b -> b
 | Char c -> c <> '\000'
 | Int i -> i <> 0
 | Float x -> x <> 0.0
 | Pointer p -> p <> PointerNil
 | Void
 | Function _
 | Struct _ ->
      raise (EvalException (pos, NotAScalar))

(*
 * Coerce to a char type.
 *)
let coerce_char pos = function
   Bool true -> '\001'
 | Bool false -> '\000'
 | Char c -> c
 | Int i -> Char.chr (i mod 256)
 | Float x -> Char.chr ((int_of_float x) mod 256)
 | Void
 | Function _
 | Struct _
 | Pointer _ ->
      raise (EvalException (pos, NotAScalar))

(*
 * Coerce to a int type.
 *)
let coerce_int pos = function
   Bool true -> 1
 | Bool false -> 0
 | Char c -> Char.code c
 | Int i -> i
 | Float x -> int_of_float x
 | Void
 | Function _
 | Struct _
 | Pointer _ ->
      raise (EvalException (pos, NotAScalar))

(*
 * Coerce to a float type.
 *)
let coerce_float pos = function
   Bool b -> if b then 1.0 else 0.0
 | Char c -> float_of_int (Char.code c)
 | Int i -> float_of_int i
 | Float x -> x
 | Void
 | Function _
 | Struct _
 | Pointer _ ->
      raise (EvalException (pos, NotAScalar))

(*
 * Get the pointer type.
 *)
let coerce_pointer pos = function
   Pointer p -> p
 | _ -> raise (EvalException (pos, NotAPointer))

(*
 * Coerce two values so they have the same type.
 *)
let coerce_equal pos v1 v2 =
   let index_of_value = function
      Bool _ -> 1
    | Char _ -> 2
    | Int _ -> 3
    | Float _ -> 4
    | Function _
    | Struct _
    | Pointer _
    | Void -> 0
   in
   let i1 = index_of_value v1 in
   let i2 = index_of_value v2 in
      if i1 = i2 then
         v1, v2
      else if i1 = 0 || i2 = 0 then
         raise (EvalException (pos, TypeError "type mismatch"))
      else if i1 > i2 then
         let v2 =
            match v1 with
               Char _ ->
                  Char (coerce_char pos v2)
             | Int _ ->
                  Int (coerce_int pos v2)
             | Float _ ->
                  Float (coerce_float pos v2)
             | _ ->
                  raise (Invalid_argument "coerce_equal")
         in
            v1, v2
      else
         let v1 =
            match v2 with
               Char _ ->
                  Char (coerce_char pos v1)
             | Int _ ->
                  Int (coerce_int pos v1)
             | Float _ ->
                  Float (coerce_float pos v1)
             | _ ->
                  raise (Invalid_argument "coerce_equal")
         in
            v1, v2

(************************************************************************
 * INITIAL ENVIRONMENT
 ************************************************************************)

(*
 * Builtin position.
 *)
let initial_pos = "<builtin>", 1, 0, 1, 0

(*
 * Initial types.
 *)
let initial_types =
   ["bool",  TypeBool initial_pos;
    "char",  TypeChar initial_pos;
    "int",   TypeInt initial_pos;
    "float", TypeFloat initial_pos]

let initial_tenv =
   let empty = SymbolTable.create () in
      List.fold_left (fun table (s, ty) ->
            SymbolTable.add table (Symbol.add s) ty) empty initial_types

(*
 * Initial values.
 *)
let initial_vals =
   ["print_string", TypeFun ([TypeArray (TypeChar initial_pos, None, initial_pos)],
                             TypeInt initial_pos,
                             initial_pos), (**)
       Function (fun pos args ->
             match args with
                [Pointer (PointerString (s, i))] ->
                   print_string (String.sub s i (String.length s - i));
                   Void
              | _ ->
                   raise (EvalException (pos, TypeError "print_string")));
    "print_int", TypeFun ([TypeInt initial_pos],
                           TypeInt initial_pos,
                           initial_pos), (**)
       Function (fun pos args ->
             match args with
                [e] ->
                   print_int (coerce_int pos e);
                   Void
              | _ ->
                   raise (EvalException (pos, TypeError "print_int")));
    "atoi", TypeFun ([TypeArray (TypeChar initial_pos, None, initial_pos)],
                     TypeInt initial_pos,
                     initial_pos), (**)
       Function (fun pos args ->
             match args with
                [Pointer (PointerString (s, i))] ->
                   Int (int_of_string (String.sub s i (String.length s - i)))
              | _ ->
                   raise (EvalException (pos, TypeError "print_string")))]

let initial_venv =
   let empty = SymbolTable.create () in
      List.fold_left (fun table (s, ty, e) ->
            SymbolTable.add table (Symbol.add s) { var_value = e; var_type = ty }) empty initial_vals

let empty_env =
   { env_types = initial_tenv;
     env_vars = initial_venv
   }

(************************************************************************
 * OPERATIONS ON VALUES
 ************************************************************************)

(*
 * Lookup a variable.
 *)
let eval_var env pos v =
   env, (lookup_var env pos v).var_value

(*
 * Pointer dereferencing.
 *)
let eval_deref pos = function
   PointerNil ->
      raise (EvalException (pos, NilPointer))
 | PointerString (s, i) ->
      if i = String.length s then
         Char '\000'
      else
         Char s.[i]
 | PointerArray (a, i) ->
      a.(i)
 | PointerVar v ->
      v.var_value

(*
 * Set the value pointed to.
 *)
let eval_set_deref pos p v2 =
   match p with
      PointerNil ->
         raise (EvalException (pos, NilPointer))
    | PointerString (s, i) ->
         s.[i] <- coerce_char pos v2
    | PointerArray (a, i) ->
         a.(i) <- v2
    | PointerVar v ->
         v.var_value <- v2

(*
 * Subscript operation.
 *)
let eval_subscript env pos v1 v2 =
   let i = coerce_int pos v2 in
   let v =
      match coerce_pointer pos v1 with
         PointerNil ->
            raise (EvalException (pos, TypeError "dereferncing nil pointer"))
       | PointerArray (a, j) ->
            let k = i + j in
               if k < 0 || k >= Array.length a then
                  raise (EvalException (pos, SubscriptOutOfBounds k));
               a.(k)
       | PointerString (s, j) ->
            let k = i + j in
            let len = String.length s in
               if k < 0 || k > len then
                  raise (EvalException (pos, SubscriptOutOfBounds k));
               if k = len then
                  Char '\000'
               else
                  Char s.[k]
       | PointerVar v ->
            if i <> 0 then
               raise (EvalException (pos, SubscriptOutOfBounds i));
            v.var_value
   in
      env, v

(*
 * Struct projection.
 *)
let eval_proj env pos v1 v2 =
   let v =
      match v1 with
         Struct s ->
            (lookup_struct pos s v2).var_value
       | _ ->
            raise (EvalException (pos, TypeError "not a struct"))
   in
      env, v

(*
 * Apply a function.
 *)
let eval_apply env pos f args =
   match f with
      Function f ->
         let v = f pos args in
            env, v
    | _ ->
         raise (EvalException (pos, TypeError "apply"))

(*
 * Return statement throws control to the
 * nearest enclosing function.
 *)
let eval_return pos v =
   raise (Return (pos, v))

let eval_break env pos =
   raise (Break pos)

(*
 * Type declaration.
 *)
let rec eval_type_decls env = function
   (v, ty, _) :: decls ->
      eval_type_decls (add_type env v ty) decls
 | [] ->
      env

(************************************************************************
 * UNARY VALUE OPERATIONS
 ************************************************************************)

(*
 * General unary arithmetic.
 *)
let eval_succ pos = function
   Int i ->
      Int (succ i)
 | Float x ->
      Float (succ_float x)
 | Char c ->
      Char (succ_char c)
 | Pointer PointerNil ->
      raise (EvalException (pos, NilPointer))
 | Pointer (PointerString (s, i)) ->
      if i >= String.length s then
         raise (EvalException (pos, SubscriptOutOfBounds (succ i)));
      Pointer (PointerString (s, succ i))
 | Pointer (PointerArray (a, i)) ->
      let i = succ i in
         if i = Array.length a then
            raise (EvalException (pos, SubscriptOutOfBounds (succ i)));
         Pointer (PointerArray (a, i))
 | Pointer (PointerVar _) ->
      raise (EvalException (pos, SubscriptOutOfBounds 1))
 | _ ->
      raise (EvalException (pos, NotAScalar))

let eval_pred pos = function
   Int i ->
      Int (pred i)
 | Float x ->
      Float (pred_float x)
 | Char c ->
      Char (pred_char c)
 | Pointer PointerNil ->
      raise (EvalException (pos, NilPointer))
 | Pointer (PointerString (s, i)) ->
      if i = 0 then
         raise (EvalException (pos, SubscriptOutOfBounds (pred i)));
      Pointer (PointerString (s, pred i))
 | Pointer (PointerArray (a, i)) ->
      if i = 0 then
         raise (EvalException (pos, SubscriptOutOfBounds (pred i)));
      Pointer (PointerArray (a, i))
 | Pointer (PointerVar _) ->
      raise (EvalException (pos, SubscriptOutOfBounds (-1)))
 | _ ->
      raise (EvalException (pos, NotAScalar))

(*
 * Uarith operates on pointers.
 *)
let eval_uarith_op pos op p =
   let pre_v = eval_deref pos p in
      match op with
         PreIncrOp ->
            let v = eval_succ pos pre_v in
               eval_set_deref pos p v;
               v
       | PreDecrOp ->
            let v = eval_pred pos pre_v in
               eval_set_deref pos p v;
               v
       | PostIncrOp ->
            eval_set_deref pos p (eval_succ pos pre_v);
            pre_v
       | PostDecrOp ->
            eval_set_deref pos p (eval_pred pos pre_v);
            pre_v

(*
 * Arithmetic negation.
 *)
let eval_uminus env pos v =
   let v =
      match v with
         Bool b ->
            Int (if b then 0 else -1)
       | Char c ->
            Int (-(Char.code c))
       | Int i ->
            Int (-i)
       | Float x ->
            Float (-.x)
       | _ ->
            raise (EvalException (pos, TypeError "uminus"))
   in
      env, v

(*
 * Arithmetic positiveness.
 *)
let eval_uplus env pos v =
   let v =
      match v with
         Bool b ->
            Int (if b then 1 else 0)
       | Char c ->
            Int (Char.code c)
       | Int i ->
            Int i
       | Float x ->
            Float x
       | _ ->
            raise (EvalException (pos, TypeError "uplus"))
   in
      env, v

(*
 * Boolean negation.
 *)
let eval_unot env pos v =
   let b =
      match v with
         Void
       | Struct _ ->
            raise (EvalException (pos, NotAScalar))
       | Int i ->
            i = 0
       | Float x ->
            x = 0.0
       | Bool b ->
            not b
       | Char c ->
            c = '\000'
       | Function _ ->
            false
       | Pointer PointerNil ->
            true
       | Pointer (PointerArray _)
       | Pointer (PointerString _)
       | Pointer (PointerVar _) ->
            false
   in
      env, Bool b

(*
 * Dereference.
 *)
let eval_ustar env pos v =
   match v with
      Pointer p ->
         env, eval_deref pos p
    | _ ->
         raise (EvalException (pos, TypeError "not a pointer"))

(*
 * Address computation.
 *)
let eval_addr_of_sub env pos v1 v2 =
   let i = coerce_int pos v2 in
   let p = coerce_pointer pos v1 in
   let v =
      if i = 0 then
         p
      else
         match p with
            PointerNil ->
               raise (EvalException (pos, NilPointer))
          | PointerString (s, j) ->
               let k = i + j in
                  if k < 0 || k > String.length s then
                     raise (EvalException (pos, SubscriptOutOfBounds k));
                  PointerString (s, k)
          | PointerArray (a, j) ->
               let k = i + j in
                  if k < 0 || k >= Array.length a then
                     raise (EvalException (pos, SubscriptOutOfBounds k));
                  PointerArray (a, k)
          | PointerVar _ ->
               raise (EvalException (pos, TypeError "addr_of"))
   in
      env, Pointer v

let eval_addr_of_proj env pos v1 v2 =
   match v1 with
      Struct s ->
         env, Pointer (PointerVar (lookup_struct pos s v2))
    | _ ->
         raise (EvalException (pos, TypeError "proj"))

(************************************************************************
 * BINARY VALUE OPERATIONS
 ************************************************************************)

(*
 * Arithmetic on ints or floats.
 *)
let eval_arith op_int op_float name env pos v1 v2 =
   let v =
      match coerce_equal pos v1 v2 with
         Int i1, Int i2 ->
            Int (op_int i1 i2)
       | Float x1, Float x2 ->
            Float (op_float x1 x2)
       | _ ->
            raise (EvalException (pos, TypeError name))
   in
      env, v

(*
 * Integer-only arith.
 *)
let eval_int_arith op name env pos v1 v2 =
   let i1 = coerce_int pos v1 in
   let i2 = coerce_int pos v2 in
      env, Int (op i1 i2)

(*
 * Comparsion arith.
 *)
let eval_bool_arith op_int op_float name env pos v1 v2 =
   let b =
      match coerce_equal pos v1 v2 with
         Int i1, Int i2 ->
            op_int i1 i2
       | Float x1, Float x2 ->
            op_float x1 x2
       | _ ->
            raise (EvalException (pos, TypeError "arith"))
   in
      env, Bool b

(*
 * Equality.
 *)
let eval_equal' env pos v1 v2 =
   match coerce_equal pos v1 v2 with
      Bool b1, Bool b2 ->
         b1 = b2
    | Char c1, Char c2 ->
         c1 = c2
    | Int i1, Int i2 ->
         i1 = i2
    | Float i1, Float i2 ->
         i1 = i2
    | Pointer PointerNil, Pointer PointerNil ->
         true
    | Pointer (PointerString (s1, i1)), Pointer (PointerString (s2, i2)) ->
         s1 == s2 && i1 = i2
    | Pointer (PointerArray (a1, i1)), Pointer (PointerArray (a2, i2)) ->
         a1 == a2 && i1 = i2
    | Pointer (PointerVar v1), Pointer (PointerVar v2) ->
         v1 == v2
    | _ ->
         raise (EvalException (pos, TypeError "equal"))

let eval_equal env pos v1 v2 =
   env, Bool (eval_equal' env pos v1 v2)

let eval_not_equal env pos v1 v2 =
   env, Bool (not (eval_equal' env pos v1 v2))

(************************************************************************
 * CORE RECURSIVE EVALUATOR
 ************************************************************************)

(*
 * Logical and.
 *)
let rec eval_land env pos v1 e2 =
   if coerce_bool pos v1 then
      eval env e2
   else
      env, v1

and eval_lor env pos v1 e2 =
   if coerce_bool pos v1 then
      env, v1
   else
      eval env e2

and eval_boolop env pos op v1 e2 =
   match op with
      LAndOp ->
         eval_land env pos v1 e2
    | LOrOp ->
         eval_lor env pos v1 e2

(*
 * Default value in a type.
 *)
and default_value env = function
   TypeBool _ ->
      Bool false
 | TypeChar _ ->
      Char '\000'
 | TypeInt _ ->
      Int 0
 | TypeFloat _ ->
      Float 0.0
 | TypeId (v, pos) ->
      default_value env (lookup_type env pos v)
 | TypeArray (ty, Some e, pos) ->
      (match eval env e with
          _, Int i ->
             if i < 0 then
                raise (EvalException (pos, TypeError "array length is negative"));
             Pointer (PointerArray (Array.init i (fun _ -> default_value env ty), 0))
        | _ ->
             raise (EvalException (pos, TypeError "illegal initializer")))
 | TypeArray _
 | TypeFun _ ->
      Pointer PointerNil
 | TypeStruct (fields, _) ->
      Struct (List.map (fun (v, ty, _) ->
                    let v' = default_value env ty in
                    let v' = { var_value = v'; var_type = ty } in
                       v, v') fields)

(*
 * Address computation.
 *)
and eval_addr_of env = function
   VarExpr (v, pos) ->
      let v = lookup_var env pos v in
         env, Pointer (PointerVar v)
 | UnOpExpr (UStarOp, e, pos) ->
      eval env e
 | SubscriptExpr (e1, e2, pos) ->
      let env, v2 = eval env e2 in
      let env, v1 = eval env e1 in
         eval_addr_of_sub env pos v1 v2
 | ProjectExpr (e, v2, pos) ->
      let env, v1 = eval env e in
         eval_addr_of_proj env pos v1 v2
 | e ->
      raise (EvalException (pos_of_expr e, TypeError "addr_of"))

(*
 * Arithmetic.
 *)
and eval_uarith env pos op e =
   let env, p = eval_addr_of env e in
   let p = coerce_pointer pos p in
   let v = eval_uarith_op pos op p in
      env, v

(*
 * General unary operations.
 *)
and eval_unop env pos op v =
   match op with
      UMinusOp ->
         eval_uminus env pos v
    | UNotOp ->
         eval_unot env pos v
    | UStarOp ->
         eval_ustar env pos v
    | UPlusOp ->
         eval_uplus env pos v
    | UBAndOp ->
         raise (Failure "& not implemented")

(*
 * General binary operations.
 *)
and eval_binop env pos op v1 v2 =
   match op with
      PlusOp ->
         eval_arith (+) (+.) "+" env pos v1 v2
    | MinusOp ->
         eval_arith (-) (-.) "-" env pos v1 v2
    | TimesOp ->
         eval_arith ( * ) ( *. ) "*" env pos v1 v2
    | DivideOp ->
         eval_arith (/) (/.) "/" env pos v1 v2
    | ModOp ->
         eval_int_arith (mod) "%" env pos v1 v2
    | BAndOp ->
         eval_int_arith (land) "&" env pos v1 v2
    | BOrOp ->
         eval_int_arith (lor) "|" env pos v1 v2
    | BXorOp ->
         eval_int_arith (lxor) "^" env pos v1 v2
    | LShiftOp ->
         eval_int_arith (lsl) "<<" env pos v1 v2
    | RShiftOp ->
         eval_int_arith (asr) ">>" env pos v1 v2
    | EqOp ->
         eval_equal env pos v1 v2
    | NotEqOp ->
         eval_not_equal env pos v1 v2
    | LeOp ->
         eval_bool_arith (<=) (<=) "<=" env pos v1 v2
    | LtOp ->
         eval_bool_arith (<) (<) "<" env pos v1 v2
    | GeOp ->
         eval_bool_arith (>=) (>=) ">=" env pos v1 v2
    | GtOp ->
         eval_bool_arith (>) (>) ">" env pos v1 v2

(*
 * Assignment.
 *)
and eval_assign env pos op e1 e2 =
   let env, p = eval_addr_of env e1 in
   let env, v2 = eval env e2 in
   let p = coerce_pointer pos p in
   let v2 =
      match op with
         Some op ->
            let v1 = eval_deref pos p in
            let env, v2 = eval_binop env pos op v1 v2 in
               eval_set_deref pos p v2;
               v2
       | None ->
            eval_set_deref pos p v2;
            v2
   in
      env, v2

(*
 * Conditional statement.
 *)
and eval_if env pos v1 e2 e3 =
   if coerce_bool pos v1 then
      eval env e2
   else
      match e3 with
         Some e3 ->
            eval env e3
       | None ->
            env, Void

(*
 * For loop.
 *)
and eval_for' env pos e2 e3 e4 =
   let env, v2 = eval env e2 in
      if coerce_bool pos v2 then
         let env, _ = eval env e4 in
         let env, _ = eval env e3 in
            eval_for' env pos e2 e3 e4
      else
         env, Void

and eval_for env pos e2 e3 e4 =
   try eval_for' env pos e2 e3 e4 with
      Break _ ->
         env, Void

(*
 * While loop.
 *)
and eval_while' env pos e1 e2 =
   let env, v1 = eval env e1 in
      if coerce_bool pos v1 then
         let env, _ = eval env e2 in
            eval_while' env pos e1 e2
      else
         env, Void

and eval_while env pos e1 e2 =
   try eval_while' env pos e1 e2 with
      Break _ ->
         env, Void

(*
 * Sequential block.
 * Drop the nested environment.
 *)
and eval_prog1 env exprs =
   match exprs with
      e :: exprs ->
         let env, v = eval env e in
         let rec loop env = function
            e :: exprs ->
               let env, _ = eval env e in
                  loop env exprs
          | [] ->
               v
         in
            env, loop env exprs
    | [] ->
         env, Void

and eval_progn env exprs =
   let rec loop env = function
      [e] ->
         eval env e
    | e :: exprs ->
         let env, _ = eval env e in
            loop env exprs
    | [] ->
         env, Void
   in
   let _, v = loop env exprs in
      env, v

(*
 * Variable declaration.
 *)
and eval_var_decls env = function
   (v, ty, e, pos) :: decls ->
      let env, v1 =
         match e with
            None ->
               env, default_value env ty
          | Some e ->
               eval env e
      in
      let env = add_var env v v1 ty in
         eval_var_decls env decls
 | [] ->
      env

(*
 * Function declaration.
 * Capture the environment (lexical scoping).
 * Also catch return values.
 *)
and add_args env pos vars args =
   match vars, args with
      (v1, ty, _) :: vars, a :: args ->
         add_args (add_var env v1 a ty) pos vars args
    | [], [] ->
         env
    | _ ->
         raise (EvalException (pos, ArityMismatch))

and eval_fun_def env pos f vars ty body =
   let ty = TypeFun (List.map (fun (_, ty, _) -> ty) vars, ty, pos) in
   let env = add_var env f (default_value env ty) ty in
   let f' pos args =
      (* Add arguments to environment *)
      let env = add_args env pos vars args in

         (* Evaluate and return the body *)
         try snd (eval env body) with
            Return (_, v) ->
               v
          | Break pos ->
               raise (EvalException (pos, TypeError "uncaught break"))
   in
      set_var env pos f (Function f'), Void

and eval_fun_decl env pos f tyl ty =
   let vars =
      let rec loop i = function
         ty :: tyl ->
            let v = Symbol.add ("v" ^ string_of_int i) in
               (v, ty, pos) :: loop (succ i) tyl
       | [] ->
            []
      in
         loop 0 tyl
   in
      eval_fun_def env pos f vars ty (SeqExpr ([], pos))

(*
 * Evaluate an expression.
 *)
and eval env = function
   IntExpr (i, _) ->
      env, Int i
 | FloatExpr (x, _) ->
      env, Float x
 | CharExpr (c, _) ->
      env, Char c
 | StringExpr (s, _) ->
      env, Pointer (PointerString (s, 0))
 | VarExpr (v, pos) ->
      eval_var env pos v
 | AddrOfExpr (e, _) ->
      eval_addr_of env e
 | UArithExpr (op, e, pos) ->
      eval_uarith env pos op e
 | UnOpExpr (op, e, pos) ->
      let env, v = eval env e in
         eval_unop env pos op v
 | BoolOpExpr (op, e1, e2, pos) ->
      let env, v1 = eval env e1 in
         eval_boolop env pos op v1 e2
 | BinOpExpr (op, e1, e2, pos) ->
      let env, v1 = eval env e1 in
      let env, v2 = eval env e2 in
         eval_binop env pos op v1 v2
 | SubscriptExpr (e1, e2, pos) ->
      let env, v1 = eval env e1 in
      let env, v2 = eval env e2 in
         eval_subscript env pos v1 v2
 | ProjectExpr (e, v, pos) ->
      let env, v1 = eval env e in
         eval_proj env pos v1 v
 | ApplyExpr (f, args, pos) ->
      let env, f = eval env f in
      let env, args = eval_list env args in
         eval_apply env pos f args
 | AssignExpr (op, e1, e2, pos) ->
      eval_assign env pos op e1 e2
 | IfExpr (e1, e2, e3, pos) ->
      let env, v1 = eval env e1 in
         eval_if env pos v1 e2 e3
 | ForExpr (e1, e2, e3, e4, pos) ->
      let env, _ = eval env e1 in
         eval_for env pos e2 e3 e4
 | WhileExpr (e1, e2, pos) ->
      eval_while env pos e1 e2
 | SeqExpr (el, _) ->
      eval_progn env el
 | ReturnExpr (e, pos) ->
      let _, v = eval env e in
         eval_return pos v
 | BreakExpr pos ->
      eval_break env pos
 | VarDefs (decls, _) ->
      eval_var_decls env decls, Void
 | FunDecl (f, args, ty, pos) ->
      eval_fun_decl env pos f args ty
 | FunDef (f, vars, ty, body, pos) ->
      eval_fun_def env pos f vars ty body
 | TypeDefs (decls, _) ->
      eval_type_decls env decls, Void

(*
 * Evaluate a list of arguments.
 *)
and eval_list env = function
   expr :: tl ->
      let env, v = eval env expr in
      let env, tl = eval_list env tl in
         env, v :: tl
 | [] ->
      env, []

(*
 * Program evaluation.
 *)
let eval_prog env prog =
   fst (eval_list env prog)

(*
 * Constant expressions.
 *)
let eval_int expr =
   let _, i = eval empty_env expr in
      match i with
         Int j -> j
       | _ -> raise (EvalException (pos_of_expr expr, NotAnInt))

(*
 * Call the main function.
 *)
let main_sym = Symbol.add "main"

let eval_main env argv =
   let pos = "<eval>", 1, 0, 1, 0 in
   let len = Array.length argv in
   let args =
      [Int (Array.length argv);
       Pointer (PointerArray (Array.map (fun s ->
                                    Pointer (PointerString (s, 0))) argv, 0))]
   in
   let _, i = eval_apply env pos (lookup_var env pos main_sym).var_value args in
   let i = coerce_int pos i in
      exit i


(*!
 * @docoff
 *
 * -*-
 * Local Variables:
 * Caml-master: "compile"
 * End:
 * -*-
 *)
