(*
 * This implementation of a graph is based on functional
 * tables.  The graph is a table of nodes, where each node
 * has a list of its immediate neighbors.
 *
 * ----------------------------------------------------------------
 *
 * Copyright (C) 1999 Jason Hickey, Caltech
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * Author: Jason Hickey
 * jyh@cs.caltech.edu
 *)

open Set_sig
open Graph_sig

(*
 * BUG: in these implementations: the nodes are represented by
 * numbers, and node allocation always increments the index.  That means
 * that for graphs with a lot of activity, the node index may overflow.
 *
 * A smarter way would be to use a real free list to manage
 * allocation of nodes.  Even better, the nodes should use weak
 * pointers, so we can reclaim parts of the graph after garbage
 * collection.
 *
 * For the compiler, these implementations should be adequate because
 * the graphs are short-lived.
 *)
module UndirectedGraph =
struct
   (*
    * A node is just a number.
    *)
   type node = int

   module NodeCompare =
   struct
      type elt = node
      let compare = (-)
   end

   module Set = Red_black_set.MakeSet (NodeCompare)
   module Table = Red_black_table.MakeTable (NodeCompare)

   (*
    * The graph contains:
    *    1. a Table that maps nodes to lists of neighbors,
    *    2. a list of nodes that are hidden
    *    3. a current allocation index for the next node.
    *)
   type t =
      { table : node Table.t;
        hidden : Set.t;
        next : node
      }

   (*
    * List operations as sets.
    *)
   let list_add x l =
      if List.mem x l then
         l
      else
         x :: l

   let list_union l1 l2 =
      List.fold_left (fun l x -> list_add x l) l1 l2

   let table_add table node1 node2 =
      Table.filter table node1 (list_add node2)

   (*
    * Create a new, empty graph.
    *)
   let create () =
      { table = Table.create ();
        hidden = Set.empty;
        next = 0
      }

   (*
    * Nodes are compared as integers.
    *)
   let eq_node n1 n2 =
      n1 = n2

   (*
    * Add a node.
    *)
   let new_node { table = table; hidden = hidden; next = next } =
      { table = Table.add_empty table next;
        hidden = hidden;
        next = succ next
      }, next

   (*
    * When an edge is added, it should be added to the
    * adjacency lists of both nodes.
    *)
   let new_edge ({ table = table } as t) node1 node2 =
      let table = Table.add table node1 node2 in
      let table = Table.add table node2 node1 in
         { t with table = table }

   (*
    * Return the list of neighbors of the node.
    *)
   let neighbors { table = table; hidden = hidden } node =
      List.filter (fun node -> not (Set.mem node hidden)) (Table.find_all table node)

   let degree { table = table; hidden = hidden } node =
      let rec count_visible i = function
         node :: nodes ->
            let i = if Set.mem node hidden then i else succ i in
               count_visible i nodes
       | [] ->
            i
      in
         count_visible 0 (Table.find_all table node)

   (*
    * Check if an edge exists.
    *)
   let query { table = table } node1 node2 =
      List.mem node2 (Table.find_all table node1)

   (*
    * When a node is deleted, all the edges to the node
    * must also be deleted.
    *)
   let delete_node { table = table; hidden = hidden; next = next } node =
      let delete_node table key =
         Table.filter table key (List_util.remove node)
      in
         { table = Table.remove (List.fold_left delete_node table (Table.find_all table node)) node;
           hidden = Set.remove node hidden;
           next = next
         }

   (*
    * Delete a single edge.
    *)
   let delete_edge ({ table = table } as t) node1 node2 =
      let table = Table.filter table node1 (List_util.remove node2) in
      let table = Table.filter table node2 (List_util.remove node1) in
         { t with table = table }

   (*
    * Map a function over the nodes.
    *)
   let fold f arg { table = table; hidden = hidden } =
      let fold_fun arg node _ =
         if Set.mem node hidden then
            arg
         else
            f arg node
      in
         Table.fold fold_fun arg table

   (*
    * Temporary hiding.
    *)
   let hide_node ({ hidden = hidden } as t) node =
      { t with hidden = Set.add node hidden }

   let restore_node ({ hidden = hidden } as t) node =
      { t with hidden = Set.remove node hidden }

   (*
    * When nodes are coalesced, only the first node remains
    * in the graph, with edges for the union of the edges of
    * both nodes.
    *)
   let coalesce ({ table = table; hidden = hidden } as t) node1 node2 =
      (* Neither node should be hidden *)
      if Set.mem node1 hidden || Set.mem node2 hidden then
         raise Not_found;

      (* Combine the edge lists *)
      let nodes2 = Table.find_all table node2 in
      let table = Table.filter table node1 (fun nodes -> list_union nodes nodes2) in
      let table = List.fold_left (fun table node2 -> table_add table node2 node1) table nodes2 in

         (* Delete the second node *)
         delete_node { t with table = table } node2
end

(*
 * The directed graph is similar to the undirected graph,
 * but we keep two values for each node: the list of predecessors
 * and the list of successors.
 *)
module DirectedGraph =
struct
   (*
    * A node is just a number.
    *)
   type node = int

   module NodeCompare =
   struct
      type elt = node
      let compare = (-)
   end

   module Set = Red_black_set.MakeSet (NodeCompare)
   module Table = Red_black_table.MakeTable (NodeCompare)

   (*
    * The graph is a Table that maps nodes to lists of neighbors,
    * and a current allocation index for the next node.
    *)
   type edges =
      { pred : Set.t;
        succ : Set.t
      }

   type t =
      { table : edges Table.t;
        next : node
      }

   (*
    * Create a new, empty graph.
    *)
   let create () =
      { table = Table.create ();
        next = 0
      }

   (*
    * Nodes are compared as integers.
    *)
   let eq_node n1 n2 =
      n1 = n2

   (*
    * Add a node.
    *)
   let new_node { table = table; next = next } =
      let node = next in
      let table = Table.add table node { pred = Set.empty; succ = Set.empty } in
         { table = table; next = succ next }, node

   (*
    * Update the value in the table.
    *)
   let update table node f =
      let f' = function
         [x] -> [f x]
       | _ -> raise (Invalid_argument "Table_graph.update")
      in
         Table.filter table node f'

   (*
    * When an edge is added, it should be added to the
    * adjacency lists of both nodes.
    *)
   let new_edge { table = table; next = next } node1 node2 =
      let add_succ { pred = pred; succ = succ } =
         { pred = pred; succ = Set.add node2 succ }
      in
      let add_pred { pred = pred; succ = succ } =
         { pred = Set.add node1 pred; succ = succ }
      in
      let table = update table node1 add_succ in
      let table = update table node2 add_pred in
         { table = table; next = next }

   (*
    * Return the list of neighbors of the node.
    *)
   let pred { table = table } node =
      Set.to_list (Table.find table node).pred

   let succ { table = table } node =
      Set.to_list (Table.find table node).succ

   (*
    * Check if an edge exists.
    *)
   let query { table = table } node1 node2 =
      Set.mem node2 (Table.find table node1).succ

   (*
    * When a node is deleted, all the edges to the node
    * must also be deleted.
    *)
   let delete_node { table = table; next = next } node =
      let remove_pred { pred = pred; succ = succ } =
         { pred = Set.remove node pred; succ = succ }
      in
      let remove_pred' table node =
         update table node remove_pred
      in
      let remove_succ { pred = pred; succ = succ } =
         { pred = pred; succ = Set.remove node succ }
      in
      let remove_succ' table node =
         update table node remove_succ
      in
      let { pred = pred; succ = succ } = Table.find table node in
      let table = List.fold_left remove_succ' table (Set.to_list pred) in
      let table = List.fold_left remove_pred' table (Set.to_list succ) in
         { table = table; next = next }

   (*
    * Delete a single edge.
    *)
   let delete_edge { table = table; next = next } node1 node2 =
      let remove_succ { pred = pred; succ = succ } =
         { pred = pred; succ = Set.remove node2 succ }
      in
      let remove_pred { pred = pred; succ = succ } =
         { pred = Set.remove node1 pred; succ = succ }
      in
      let table = update table node1 remove_succ in
      let table = update table node2 remove_pred in
         { table = table; next = next }

   (*
    * In the depth-first-search, the graph is traversed
    * from the originating node outward.  We keep a bit-vector
    * for marking nodes as we traverse them.
    *)
   let depth_first_search { table = table; next = next } node f arg1 arg2 =
      let marked = Array.create next false in
      let rec search arg1 arg2 = function
         node :: nodes ->
            let arg1 =
               if marked.(node) then
                  arg1
               else
                  let arg1, arg2 = f arg1 arg2 node in
                     marked.(node) <- true;
                     search arg1 arg2 (Set.to_list (Table.find table node).succ)
            in
               search arg1 arg2 nodes
       | [] ->
            arg1
      in
         search arg1 arg2 [node]
end

(*
 * -*-
 * Local Variables:
 * Caml-master: "compile"
 * End:
 * -*-
 *)
